﻿/**
HTTP-easyDown

HTTP通信でファイルを取得する実験プログラム。

<table>
	<caption>おぷしょん。</caption>
	<thead>
		<tr>
			<th>スイッチ</th>
			<th>オプション</th>
			<th>備考</th>
		</tr>
	</thead>
	<tbody>
		<tr>
			<td>host</td>
			<td>ホスト名</td>
			<td>IPアドレスもしくはホスト名を指定。IPv6未対応。</td>
		</tr>
		<tr>
			<td>path</td>
			<td>ホストに要求するパス</td>
			<td>未指定の場合はサーバ依存。</td>
		</tr>
		<tr>
			<td>port</td>
			<td>接続先ポート番号</td>
			<td>未指定の場合は80を使用。</td>
		</tr>
		<tr>
			<td>user</td>
			<td>接続ユーザー名</td>
			<td>必要ならば。</td>
		</tr>
		<tr>
			<td>pass</td>
			<td>ユーザー名に対応するパスワード</td>
			<td>必要ならば。</td>
		</tr>
		<tr>
			<td>ua</td>
			<td>User-Agent</td>
			<td>ユーザーエージェント。未指定ならば<strong>HTTP-easyDown/<em>version</em></strong></td>
		</tr>
		<tr>
			<td>lang</td>
			<td>Accept-Language</td>
			<td>必要ならば。</td>
		</tr>
		<tr>
			<td>s_head</td>
			<td>標準出力へヘッダ書き出し</td>
			<td>未指定なら出力されない。</td>
		</tr>
		<tr>
			<td>s_body</td>
			<td>標準出力へボディ書き出し</td>
			<td>未指定なら出力されない。</td>
		</tr>
		<tr>
			<td>f_head</td>
			<td>ファイルへヘッダ書き出し</td>
			<td>未指定なら出力されない。</td>
		</tr>
		<tr>
			<td>f_body</td>
			<td>ファイルへボディ書き出し</td>
			<td>未指定なら出力されない。</td>
		</tr>
		<tr>
			<td>output</td>
			<td>出力ファイル</td>
			<td>f_headとf_bodyのどちらかが指定されていた場合に有効。</td>
		</tr>
		<tr>
			<td>retry</td>
			<td>再接続回数</td>
			<td>未指定なら0回</td>
		</tr>
		<tr>
			<td>wait</td>
			<td>待機時間(ms)</td>
			<td>取得停止から待機する時間。未指定なら500ms</td>
		</tr>
		<tr>
			<td>size</td>
			<td>Range</td>
			<td>取得サイズ。</td>
		</tr>
		<tr>
			<td>buf</td>
			<td>読み込みブロックサイズ</td>
			<td>指定しなければ1024byte。</td>
		</tr>
	</tbody>
</table>

<ul>
	<li>いろいろ考えては見たけれどもLocationの面倒はこっちで見ないことにした</li>
	<li>gzip展開がよく分からん。ていうか31, 139が頭に来てないところもある？ｼﾗﾈ</li>
	<li>↑Accept-Encoding: deflateで放棄</li>
</ul>
*/
module httpeasydown;

debug import std.stdio: wl = writefln, pl = printf;
debug {import std.perf; scope class Timer:PerformanceCounter{this(){start;}~this(){stop;wl("Timer:%s",microseconds);}}}

import std.getopt;
import std.socket;
import std.stdio;
import std.string;
import std.date;
import core.thread;
static import std.base64;
import std.stream, std.socketstream;
//import std.zlib;

invariant auto NAME = "HTTP-easyDown";
invariant auto VER  = "1.00";

/**
*/
int main(string[] args) {
	if(args.length == 1) {
		help();
		return 0;
	}
	
	Request request;
	request.getOptions(args);

	auto host=new InternetHost();
	if(!host.getHostByName(request.Host))
		throw new Exception("host or IP ?");
	auto addr=new InternetAddress(host.name, request.Port);

	TcpSocket tcp;
	scope(exit) if(tcp) tcp.close();
	
	for(auto i=0; i < request.ReTry+1; i++) {
		try {
			tcp=new TcpSocket(addr);
			break;
		} catch(SocketException e) {
			if(i)
				writefln("@retry %s / %s", i, request.ReTry);
		}
	}
	
	if(!tcp || !tcp.isAlive()) {
		throw new Exception("can't connect.");
	}

	auto srss=new SocketStream(tcp);
	scope(exit) if(srss) srss.close();

	auto httpReq=request.toString();
	srss.writeBlock(httpReq.ptr, httpReq.length);
	srss.flush();

	// ファイル用ストリーム。
	BufferedFile fs;
	scope(exit) if(fs) fs.close();
	if(request.FileHeader || request.FileBody) {
		fs = new BufferedFile(request.Output, FileMode.OutNew);
	}

	// 受信メモリ用ストリーム。
	MemoryStream ms;
	scope(exit) if(ms) ms.close();

	// 一回分の受信バッファ。
	invariant auto SIZE_BLOCK = request.Buffer;
	auto block=new ubyte[SIZE_BLOCK];

	// ヘッダループ
	string HttpHead;
	string[string] HeadData;
	foreach(char[] head; srss) {
		
		if(!head.length) {
			write("\r\n");
			HttpHead ~= "\r\n";
			break;
		}
		
		HttpHead ~= cast(string)head ~ "\r\n";

		int pos = find(head, ":");
		if(pos != -1) {
			HeadData[strip(head[0..pos].idup)] = strip(head[pos+1..$].idup);
		}
		
		if(request.StdHeader)
			writeln(head);
	}
	if(request.FileHeader) {
		fs.write(cast(ubyte[])HttpHead);
		fs.flush();
	}
	
	// ボディの取得を行わないのであれば終了
	if(!request.FileBody && !request.StdBody) {
		return 0;
	}
	{ //-------------------------------------------------------------------
		// 最後のLFがシークされてない？
		ubyte dummy;
		srss.read(dummy);
	} //-------------------------------------------------------------------
	
	// ボディ取得。
	ms=new MemoryStream();
	size_t AllSize;
	do {
		auto size = srss.read(block);
		AllSize += size;
		if(!srss.eof && size == 0) {
			// ( ﾟдﾟ) やっぱり寝る!
			Thread.sleep(request.WaitTime * 1000);
		}
		ms.writeBlock(block.ptr, size);
	} while(!srss.eof);
	
	/+ よくわからん
	// 圧縮されてる？
	bool gzip = cast(bool)("Content-Encoding" in HeadData) ? HeadData["Content-Encoding"] == "gzip": false;
	if(gzip) {
		auto Archive = new ubyte[AllSize];
		ms.seekSet(0);
		wl(ms.readBlock(Archive.ptr, Archive.length));
		//ms.close;
		wl("<bin>\n", Archive);
		wl("<raw>\n", ms);
		ms = new MemoryStream(cast(ubyte[])uncompress(cast(void[])Archive[8..$], AllSize));
	}
	+/
	if(request.StdBody)
		write(ms);
	
	if(request.FileBody) {
		fs.write(ms.data);
		fs.flush();
	}

	return 0;
}

/**
*/
struct Request {
	string Host;
	string Path = "/";
	ushort Port = 80;
	string User;
	string Pass;
	string UserAgent;
	string Language;
	bool StdHeader;
	bool StdBody;
	bool FileHeader;
	bool FileBody;
	string Output;
	size_t ReTry = 0;
	d_time WaitTime = 500;
	size_t Range = 0;
	size_t Buffer = 1024;
	
	/**
	*/
	void getOptions(string[] args) {
		getopt(
			args,
			config.caseSensitive,
			config.noPassThrough,
			"host"   , &Host,
			"path"   , &Path,
			"port"   , &Port,
			"user"   , &User,
			"pass"   , &Pass,
			"ua"     , &UserAgent,
			"lang"   , &Language,
			"s_head" , &StdHeader,
			"s_body" , &StdBody,
			"f_head" , &FileHeader,
			"f_body" , &FileBody,
			"output" , &Output,
			"retry"  , &ReTry,
			"wait"   , &WaitTime,
			"size"   , &Range,
			"buf"    , &Buffer
		);
		if(!Host)
			throw new Exception("options --host=?");

		foreach(FilePut; [FileHeader, FileBody]) {
			if(FilePut && !Output) {
				throw new Exception("options --output=?");
			}
		}
	}

	/**
	*/
	string toString() {
		string[] lines;

		if(!Path.length)
			Path = "/";
		if(Path[0] != '/')
			Path = '/' ~ Path;
		
		lines ~= format("GET %s HTTP/1.1", Path);
		
		lines ~= "Accept-Encoding: deflate";
		//lines ~= "Accept-Encoding: gzip";

		if(Range)
			lines ~= format("Range: bytes=0-%s", Range-1);
		
		if(Language)
			lines ~= format("Accept-Language: %s", Language);

		if(User && Pass)
			lines ~= format("Authorization: Basic %s", std.base64.encode(User ~ ":" ~ Pass));

		if(!UserAgent) {
			UserAgent = format("%s/%s", NAME, VER);
		}
		lines ~= format("User-Agent: %s", UserAgent);
		
		lines ~= format("Host: %s: %s", Host, Port);
		lines ~= "Connection: Close";
		

		return lines.join("\r\n") ~ "\r\n" ~ "\r\n";
	}
}

void help() {
	writefln("%s %s", NAME, VER);
	writeln(line);
	
	string[][] list = [
	["host=<...>",   `The host name or IP address. Not compatible with IPv6.`],
	["path=/",       `Demand a path to the host. If unspecified "/".`],
	["port=80",      `The destination port number. If you use unspecified 80.`],
	["user=<?>",     `If necessary connection users name.`],
	["pass=<?>",     `Password for the user name.`],
	["ua=<?>",       `User-Agent. If unspecified "HTTPGet/version"`],
	["lang=<?>",     `Accept-Language. If necessary`],
	["s_head",       `Export standard output to the header.`],
	["s_body",       `Export standard output to the body.`],
	["f_head",       `Export to a file header.`],
	["f_body",       `Export to a file body.`],
	["output=<...>", `"f_head" or "f_body" in the case of output file name.`],
	["retry=0",      `Number of reconnecting.`],
	["wait=500",     `Waiting time(ms).If unspecified 500ms.`],
	["size=<?>",     `Range.`],
	["buf=1024",     `Read block size. If unspecified 1024byte`]
	];

	foreach(str; list) {
		writefln("--%12-s\t%s", str[0], str[1]);
	}

	writeln(line);
	writeln("Thank you, google translation. http://translate.google.co.jp/translate_t?hl=ja#");
}

string line() {
	return repeat("-", 79);
}
